/* soli.c - Simulation von zwei Solitonen
 *
 * Im oberen Fenster wird die analytische Loesung angezeigt,
 * im unteren die numerische.
 */

#include <gtk/gtk.h>
#include <math.h>
#include <pthread.h>
#include <semaphore.h>

/* x_start und x_end legen den x-Bereich fest, der
 * angezeigt werden soll. Entsprechend legen t_start
 * und t_end den Zeitbereich fest, der simuliert
 * werden soll. t_step gibt an, in welchen Abstaenden
 * der Bildschirm aktualisiert werden soll.
 */
double x_start = -30.0;
double x_end = 30.0;
double t_start = -1.0;
double t_end = 1.8;
double t_step = 0.002;
double t;

GdkPixmap *pix_numeric = NULL;
GdkPixmap *pix_analytic = NULL;
GtkWidget *draw1 = NULL;
int width1, height1;
GtkWidget *draw2 = NULL;
int width2, height2;
sem_t draw_config_sem;

/* die analytische Loesung */
double analytic(double x, double t)
{
	double tmp;

	tmp = cosh(3.0*x-((8.0+4.0*4.0*4.0)/2.0)*t)
	      + 3.0*cosh(x+((8.0-4.0*4.0*4.0)/2.0)*t);

	return 12.0 * (3.0 + 4.0*cosh(2.0*x-8.0*t)
		       + cosh(4.0*x-(4.0*4.0*4.0)*t)
		      ) / (tmp*tmp);
}

/* zeichnen der analytischen Loesung in pix_analytic */
void draw_analytic()
{
	int i, y, last_y;
	double x, x_step, y_scale;

	y_scale = height1/9.0;
	x_step = (x_end-x_start)/width1;
	x = x_start;
	last_y = height1 - analytic(x, t)*y_scale - 1;

	gdk_draw_rectangle(pix_analytic,
			   draw1->style->white_gc,
			   TRUE, 0, 0, width1, height1);

	for (i=0; i<width1; i++) {
		x += x_step;
		y = height1 - analytic(x, t)*y_scale - 1;
		gdk_draw_line(pix_analytic,
			      draw1->style->black_gc,
			      i, last_y, i+1, y);
		last_y = y;
	}
}

/* Variablen fuer die numerische Berechnung */
#define xsteps 1000		/* Zahl der Punkte in x-Richtung */
double numeric_xstep;		/* Schrittweite in x-Richtung */
#define numeric_tstep 0.0000005	/* Schrittweite in t-Richtung */
double numeric_data1[xsteps];
double numeric_data2[xsteps];
double *numeric_data = numeric_data1;
double *last_numeric_data = numeric_data2;
double last_t;

/* Festlegen der Anfangswerte fuer die numerische Loesung */
void reset_numeric()
{
	int i;

	t = t_start;
	last_t = t_start;

	numeric_xstep = (x_end-x_start)/(double)xsteps;

	for (i=0; i<xsteps; i++) {
		numeric_data[i] = analytic(x_start
					   + numeric_xstep*i,
					   t);
		last_numeric_data[i] = numeric_data[i];
	}
}

/* numerische Berechnung der ersten Ableitung nach x */
inline double calc_d(int i)
{
	return (last_numeric_data[i-2]
		- last_numeric_data[i+2]
		+ 8.0*(last_numeric_data[i+1]
		       - last_numeric_data[i-1])
	       ) / (12.0*numeric_xstep);
}

/* numerische Berechnung der dritten Ableitung nach x */
inline double calc_d3(int i)
{
	return (last_numeric_data[i+3]
		- 3.0*last_numeric_data[i+1]
		+ 3.0*last_numeric_data[i-1]
		- last_numeric_data[i-3]
	       ) / (8.0*numeric_xstep
		    *numeric_xstep*numeric_xstep);
}

/* numerische Berechnung der Loesung */
void calc_numeric()
{
	int i;
	double dudt;
	double *tmp;

	while ((t-last_t)<t_step){
		t += numeric_tstep;
		tmp = numeric_data;
		numeric_data = last_numeric_data;
		last_numeric_data = tmp;

		for (i=3; i<xsteps-3; i++) {
			dudt = - calc_d3(i)
			       - 6.0*last_numeric_data[i]*calc_d(i);

			numeric_data[i] = last_numeric_data[i]
					  + dudt*numeric_tstep;
		}

		pthread_testcancel();
	}

	last_t = t;
}

/* zeichnen der numerischen Loesung in pix_numeric */
void draw_numeric()
{
	int i, y, last_y, xi;
	double x, x_step, y_scale;

	y_scale = height2/9.0;
	x_step = (x_end-x_start)/width2;
	x = x_start;
	last_y = height2 - analytic(x, t)*y_scale - 1;

	gdk_draw_rectangle(pix_numeric,
			   draw2->style->white_gc,
			   TRUE, 0, 0, width2, height2);

	for (i=0; i<width2; i++) {
		x += x_step;
		xi = (x-x_start)/numeric_xstep;
		y = height2 - numeric_data[xi]*y_scale - 1;
		gdk_draw_line(pix_numeric,
			      draw2->style->black_gc,
			      i, last_y, i+1, y);
		last_y = y;
	}
}

/* aktualisieren der beiden Graphen */
void update()
{
	gdk_draw_drawable(draw1->window,
			  draw1->style
			       ->fg_gc[GTK_WIDGET_STATE(draw1)],
			  pix_analytic, 0, 0, 0, 0,
			  width1, height1);
	gdk_draw_drawable(draw2->window,
			  draw2->style
			       ->fg_gc[GTK_WIDGET_STATE(draw2)],
			  pix_numeric, 0, 0, 0, 0,
			  width2, height2);
}

/* die Hauptschleife, wird bei druecken
 * des Start-buttons ausgefuehrt */
void *run(void *arg)
{
	while(t<t_end) {
		calc_numeric();

		sem_wait(&draw_config_sem);
		draw_numeric();
		draw_analytic();
		update();
		sem_post(&draw_config_sem);
	}

	return NULL;
}

/* es folgen einige Funktionen zum Abhandeln der events */

void destroy(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

gboolean configure_event1(GtkWidget *widget, GdkEventConfigure *event)
{
	sem_wait(&draw_config_sem);

	if (pix_analytic)
		g_object_unref(pix_analytic);

	width1 = draw1->allocation.width;
	height1 = draw1->allocation.height;
	pix_analytic = gdk_pixmap_new(widget->window,
				      width1, height1, -1);
	draw_analytic();

	sem_post(&draw_config_sem);

	return TRUE;
}

gboolean configure_event2(GtkWidget *widget, GdkEventConfigure *event)
{
	sem_wait(&draw_config_sem);

	if (pix_numeric)
		g_object_unref(pix_numeric);

	width2 = draw2->allocation.width;
	height2 = draw2->allocation.height;
	pix_numeric = gdk_pixmap_new(widget->window,
				     width2, height2, -1);
	draw_numeric();

	sem_post(&draw_config_sem);

	return TRUE;
}

gboolean expose_event1(GtkWidget *widget,
		       GdkEventExpose *event,
		       GdkPixmap **pixmap)
{
	gdk_draw_drawable(widget->window,
			  widget->style
				->fg_gc[GTK_WIDGET_STATE(widget)],
			  pix_analytic,
			  event->area.x, event->area.y,
			  event->area.x, event->area.y,
			  event->area.width, event->area.height);

	return FALSE;
}

gboolean expose_event2(GtkWidget *widget,
		       GdkEventExpose *event,
		       GdkPixmap **pixmap)
{
	gdk_draw_drawable(widget->window,
			  widget->style
				->fg_gc[GTK_WIDGET_STATE(widget)],
			  pix_numeric,
			  event->area.x, event->area.y,
			  event->area.x, event->area.y,
			  event->area.width, event->area.height);

	return FALSE;
}

pthread_t tr;
int running = 0;

void start(GtkWidget *widget, gpointer data)
{
	if (running)
		return;

	pthread_create(&tr, NULL, run, NULL);
	running = 1;
}

void reset(GtkWidget *widget, gpointer data)
{
	if (!running)
		return;

	pthread_cancel(tr);
	running = 0;
	reset_numeric();
	draw_numeric();
	draw_analytic();
	update();
}

/* stetzen der Anfangswerte und Aufbau des Fensters */
int main(int argc, char *argv[])
{
	GtkWidget *window, *button1, *button2, *box1, *box2, *box3;

	sem_init(&draw_config_sem, 0, 1);
	reset_numeric();
	gtk_init(&argc, &argv);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "Solitons");
	g_signal_connect(G_OBJECT(window), "destroy",
			 G_CALLBACK(destroy), NULL);
	gtk_container_set_border_width(GTK_CONTAINER(window), 10);

	box1 = gtk_hbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(window), box1);
	gtk_widget_show(box1);

	box2 = gtk_vbox_new(FALSE, 10);
	gtk_box_pack_start(GTK_BOX(box1), box2, TRUE, TRUE, 0);
	gtk_widget_show(box2);

	box3 = gtk_vbox_new(FALSE, 10);
	gtk_box_pack_start(GTK_BOX(box1), box3, TRUE, TRUE, 0);
	gtk_widget_show(box3);

	draw1 = gtk_drawing_area_new();
	gtk_widget_set_size_request(draw1, 500, 200);
	gtk_signal_connect(GTK_OBJECT(draw1), "expose_event",
			   (GtkSignalFunc) expose_event1, NULL);
	gtk_signal_connect(GTK_OBJECT(draw1), "configure_event",
			   (GtkSignalFunc) configure_event1, NULL);
	gtk_box_pack_start(GTK_BOX(box2), draw1, TRUE, TRUE, 0);
	gtk_widget_show(draw1);

	draw2 = gtk_drawing_area_new();
	gtk_widget_set_size_request(draw2, 500, 200);
	gtk_signal_connect(GTK_OBJECT(draw2), "expose_event",
			   (GtkSignalFunc) expose_event2, NULL);
	gtk_signal_connect(GTK_OBJECT(draw2), "configure_event",
			   (GtkSignalFunc) configure_event2, NULL);
	gtk_box_pack_start(GTK_BOX(box2), draw2, TRUE, TRUE, 0);
	gtk_widget_show(draw2);

	button1 = gtk_button_new_with_label("Start");
	g_signal_connect(G_OBJECT(button1), "clicked",
			 G_CALLBACK(start), NULL);
	gtk_box_pack_start(GTK_BOX(box3), button1, TRUE, TRUE, 0);
	gtk_widget_show(button1);

	button2 = gtk_button_new_with_label("Reset");
	g_signal_connect(G_OBJECT(button2), "clicked",
			 G_CALLBACK(reset), NULL);
	gtk_box_pack_start(GTK_BOX(box3), button2, TRUE, TRUE, 0);
	gtk_widget_show(button2);

	gtk_widget_show(window);
	gtk_main();

	return 0;
}
